const rabbit = require('rabbot');

// If used in Sails
const _error = typeof sails !== 'undefined' ? sails.log.error : console.error; // eslint-disable-line
const _info = typeof sails !== 'undefined' ? sails.log.info : console.info; // eslint-disable-line

const CONST_SERVER_ROLE_WEB = 'web';
const CONST_SERVER_ROLE_WORKER = 'worker';
const CONST_SERVER_MODE_MS = 'master/slave';
const CONST_SERVER_MODE_M = 'master';

const CONST_MIN_DELAY_MINUTES = 1;
const CONST_MAX_DELAY_MINUTES = 255;

module.exports = {
  roles: {
    web: CONST_SERVER_ROLE_WEB,
    worker: CONST_SERVER_ROLE_WORKER
  },

  modes: {
    m: CONST_SERVER_MODE_M,
    ms: CONST_SERVER_MODE_MS
  },

  serverRole: null,
  serverMode: null,
  silent: false,
  rabbitNotAvailable: false,

  init(
    rabbitConfig,
    {
      silent = false,
      serverRole = CONST_SERVER_ROLE_WEB,
      serverMode = CONST_SERVER_MODE_MS,
      errTracker,
      routingKeys = { 'std-jobs': 'std-jobs', mailer },
      minDelay = CONST_MIN_DELAY_MINUTES,
      maxDelay = CONST_MAX_DELAY_MINUTES
    } = {}
  ) {
    return new Promise((resolve, reject) => {
      this.serverRole = serverRole;
      this.serverMode = serverMode;
      this.silent = silent;
      this._info = _info;
      this.routingKeys = routingKeys;
      this.minDelay = minDelay;
      this.maxDelay = maxDelay;

      if (errTracker) {
        this.errorFunction = errTracker.trackError;
      } else {
        this.errorFunction = function(
          error,
          errorCode,
          payload,
          req,
          tagsToAdd
        ) {
          _error(error, errorCode, tagsToAdd);
        };
      }

      if (
        serverRole === CONST_SERVER_ROLE_WEB &&
        serverMode === CONST_SERVER_MODE_MS
      ) {
        rabbit.on('closed', () => {
          this.errorFunction('RabbitMQ closed', 500, {}, null, [
            'worker',
            'rabbitMQ',
            'on closed'
          ]);
          rabbit.retry();
        });

        rabbit.on('failed', () => {
          this.errorFunction('RabbitMQ failed', 500, {}, null, [
            'worker',
            'rabbitMQ',
            'on failed'
          ]);
          rabbit.retry();
        });

        rabbit.on('unreachable', () => {
          this.errorFunction('RabbitMQ unreachable', 500, {}, null, [
            'worker',
            'rabbitMQ',
            'on unreachable'
          ]);
        });

        rabbit
          .configure(rabbitConfig)
          .then(() => {
            this.info('RabbitMQ config successful');
            resolve();
          })
          .catch((err) => {
            this.error(err, null, {}, null, ['q', 'init']);
            _error('rabbit-configure: Could not connect to Q');
            this.rabbitNotAvailable = true;
            reject(err);
          });
      } else {
        resolve();
      }
    });
  },

  error(...args) {
    if (!this.silent) {
      this.errorFunction(args);
    }
  },

  info(...args) {
    if (!this.silent) {
      _info(args);
    }
  },

  shutdown() {
    rabbit.shutdown();
  },

  /**
   * Determines if this backend instance can delegate the work to Q or has to do the work on its own
   * @return {Boolean}
   */
  mustIWork() {
    let mustIWork = false;
    if (this.serverRole === CONST_SERVER_ROLE_WORKER) {
      mustIWork = true;
    } else if (
      this.serverMode === CONST_SERVER_MODE_M ||
      this.rabbitNotAvailable
    ) {
      mustIWork = true;
    }
    return mustIWork;
  },

  /**
   * Delegates a task to the rabbitMQ exchanges std-jobs
   * @param  {Object}   task the task description as function "location" specified by type, item, method and params, e.g.
   *
   * Example for task:
   *    const task = {
   *        type: 'services',
   *        item: 'mailer',
   *        method: '_testSend',
   *        params: []
   *    };
   *
   * @param  {Function} cb   a callback after queue publishing
   */
  task(task, cb) {
    cb = cb || function() {};

    if (
      task &&
      task.type &&
      task.item &&
      task.type === 'services' &&
      task.item === 'q'
    ) {
      return cb("Don't mess with spaceTime. You are not supposed to q the q.");
    }

    const routingKey = this.routingKeys['std-jobs'];

    rabbit
      .publish('exchange', {
        type: 'task',
        body: task,
        routingKey
      })
      .then(cb)
      .catch((err) => {
        this.error(
          err,
          500,
          { task: task, routingKey },
          null,
          ['q', 'task'],
          cb
        );
      });
  },

  /**
   * Delegates a task to the rabbitMQ exchanges std-jobs WITH a DELAY in minutes
   * DOES NOT DEBOUNCE, ONLY DELAY
   *
   * @param  {Object}   task the task description as function "location" specified by type, item, method and params, e.g.
   *
   * Example for task:
   *    const task = {
   *        type: 'services',
   *        item: 'mailer',
   *        method: '_testSend',
   *        params: []
   *    };
   *
   * @param  {Number}  delayInMins the number of minutes to delay this task,
   */
  // delayTask(task, delayInMins) {
  //   if (
  //     task &&
  //     task.type &&
  //     task.item &&
  //     task.type === 'services' &&
  //     task.item === 'q'
  //   ) {
  //     return this.error(
  //       "Don't mess with spaceTime. You are not supposed to q the q.",
  //       500,
  //       { task, delayInMins },
  //       null,
  //       ['q', 'delayTask']
  //     );
  //   }

  //   const routingKey = this._buildDelayRoutingKey(delayInMins);

  //   rabbit
  //     .publish('delay-ex', {
  //       type: 'task',
  //       body: task,
  //       routingKey
  //     })
  //     .catch((err) => {
  //       this.error(err, 500, { task, routingKey, delayInMins }, null, [
  //         'q',
  //         'delayTask'
  //       ]);
  //     });
  // },

  /**
   * Requests a function call via the message queue and responds back with the result using the callback (RPC)
   *
   * @param  {Object}   task the task description as function "location" specified by type, item, method and params, e.g.
   *
   * Example for task:
   *    const task = {
   *        type: 'services',
   *        item: 'mailer',
   *        method: '_testSend',
   *        params: []
   *    };
   *
   * @param  {Function} cb   The function to be called with the result once we get the result back from a worker
   */
  req(task, cb = () => {}) {
    const routingKey = this.routingKeys['std-jobs'];

    if (typeof task.cbPos !== 'number') {
      //debug ('assigned the last pos or args as cbPos');
      task.cbPos = task.params.length - 1;
    }
    if (
      task &&
      task.type &&
      task.item &&
      task.type === 'services' &&
      task.item === 'q'
    ) {
      return this.error(
        "Don't mess with spaceTime. You are not supposed to q the q.",
        null,
        { task: task },
        null,
        ['q', 'req'],
        cb
      );
    }

    rabbit
      .request('exchange', {
        type: 'req',
        body: task,
        routingKey
      })
      /*.progress(function(reply) {
            // if multiple replies are provided, all but the last will be sent via the progress callback
          })*/

      .then((final) => {
        // the last message in a series OR the only reply will be sent to this callback
        // debug ('final@then :', final);

        final.ack();
        if (
          Array.isArray(task.params) &&
          typeof task.params[task.cbPos] === 'function'
        ) {
          // debug ('typeof argument@final.body.args:', typeof final.body.args);
          task.params[task.cbPos].apply(this, final.body.args);
          // debug ('done?');
        } else {
          this.error(
            'Failed to call the callback supplied(or assumed) to the q.req. Please check if the cb at cbPos is a function.',
            null,
            { task: task },
            null,
            ['q', 'req', 'rabbit.request.then'],
            cb
          );
        }
      })
      .catch((err) => {
        this.error(
          err,
          500,
          { task, routingKey },
          null,
          ['q', 'req', 'rabbit.request.catch'],
          cb
        );
      });
  },

  message(eventName, message, cb = () => {}) {
    if (!eventName || !message) {
      this.error(
        'eventName or message are not correct',
        null,
        { eventName },
        null,
        ['q', 'message']
      );
    }

    const routingKey = this.routingKeys['mailer'];
    rabbit
      .publish('router', {
        type: 'message',
        body: { eventName, message },
        routingKey
      })
      .catch((err) => {
        this.error(
          err,
          500,
          { message, routingKey },
          null,
          ['q', 'req', 'rabbit.message.catch'],
          cb
        );
      });
  }

  /**
   * Builds a delay routing key based on a number of minutes of delay
   * It is crafted to comply to the topic wildcard routing keys for cascading exchanges and queues
   * It transforms the number into a binary format separated by a '.' per element, pads it with 0s up to 7 positions.
   * @param  {Number} delayInMins The number of minutes to delay the task execution
   * @return {String}             A routing key as a String complying to the delay exchanges routing key regex patterns, e.g. '1.0.0.0.0.0.0.delay' or "0.0.0.0.1.0.1.delay"
   */
  // _buildDelayRoutingKey(delayInMins = 1) {
  //   if (delayInMins > this.maxDelay) {
  //     delayInMins = this.maxDelay;
  //   }

  //   if (!delayInMins) {
  //     delayInMins = this.minDelay;
  //   }

  //   return parseInt(delayInMins, 10)
  //     .toString(2)
  //     .padStart(7, '0')
  //     .split('')
  //     .reverse()
  //     .join('.')
  //     .concat('.delay');
  // }
};
